/* quartz_database.h: C++ class definition for quartz database
 *
 * ----START-LICENCE----
 * Copyright 1999,2000,2001 BrightStation PLC
 * Copyright 2002 Ananova Ltd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * -----END-LICENCE-----
 */

#ifndef OM_HGUARD_QUARTZ_DATABASE_H
#define OM_HGUARD_QUARTZ_DATABASE_H

#include "database.h"

class QuartzTableManager;
class QuartzBufferedTableManager;
class QuartzModifications;
class QuartzTermList;

#include "autoptr.h"

/** A backend designed for efficient indexing and retrieval, using
 *  compressed posting lists and a btree storage scheme.
 */
class QuartzDatabase : public Database {
    friend class QuartzWritableDatabase;
    friend class QuartzTermList;
    private:
	/** Pointer to table manager.
	 */
	AutoPtr<QuartzTableManager> tables;

	//@{
	/** Implementation of virtual methods: see Database for details.
	 */
	virtual void do_begin_session();
	virtual void do_end_session();
	virtual void do_flush();

	virtual void do_begin_transaction();
	virtual void do_commit_transaction();
	virtual void do_cancel_transaction();

	virtual om_docid do_add_document(const OmDocument & document);
	virtual void do_delete_document(om_docid did);
	virtual void do_replace_document(om_docid did,
					 const OmDocument & document);

	virtual void do_reopen();
	//@}

	/// Implementation of get_avlength()
	om_doclength get_avlength_internal() const;

	/// Implementation of open_post_list()
	LeafPostList * open_post_list_internal(const om_termname & tname,
				RefCntPtr<const Database> ptrtothis) const;

	/// Implementation of open_term_list()
	LeafTermList * open_term_list_internal(om_docid did,
				RefCntPtr<const Database> ptrtothis) const;

	/// Implementation of get_doccount()
	om_doccount  get_doccount_internal() const;
    public:
	/** Create and open a quartz database.
	 *
	 *  @exception OmOpeningError thrown if database can't be opened.
	 *
	 *  @param params Parameters supplied by the user to specify the
	 *                location of the database to open.
	 */
	QuartzDatabase(const OmSettings & params);

	/** Constructor used by QuartzWritableDatabase.
	 *
	 *  @param tables_ A pointer to the tables to use.
	 */
	QuartzDatabase(AutoPtr<QuartzTableManager> tables_);

	~QuartzDatabase();

	/// Get the directory that the database is stored in.
	static std::string get_db_dir(const OmSettings & settings);

	/// Get the file to use as a logfile.
	static std::string get_log_filename(const OmSettings & settings);

	/// Get the block size to use.
	static unsigned int get_block_size(const OmSettings & settings);

	/// Get whether a new database should be created.
	static bool get_create(const OmSettings & settings);

	/** Get whether we should allow creation over the top of an old
	 *  database.
	 */
	static bool get_allow_overwrite(const OmSettings & settings);

	/** Virtual methods of Database.
	 */
	//@{
	om_doccount  get_doccount() const;
	om_doclength get_avlength() const;
	om_doclength get_doclength(om_docid did) const;
	om_doccount get_termfreq(const om_termname & tname) const;
	om_termcount get_collection_freq(const om_termname & tname) const;
	bool term_exists(const om_termname & tname) const;

	LeafPostList * do_open_post_list(const om_termname & tname) const;
	LeafTermList * open_term_list(om_docid did) const;
	Document * open_document(om_docid did, bool lazy = false) const;
	AutoPtr<PositionList> open_position_list(om_docid did,
					  const om_termname & tname) const;
	TermList * open_allterms() const;
	//@}
};

/** A writable quartz database.
 */
class QuartzWritableDatabase : public Database {
    private:
	/** Pointer to buffered table manager.
	 *
	 *  This points to the same object as tables, but is used for
	 *  modification access.
	 */
	QuartzBufferedTableManager * buffered_tables;

	/** A count of the number of changes since the last flush:
	 *  FIXME: this should be replaced by keeping track of the memory used
	 *  up, and flushing when it reaches a critical value.
	 */
	int changecount;

	/** The readonly database encapsulated in the writable database.
	 */
	QuartzDatabase database_ro;

	//@{
	/** Implementation of virtual methods: see Database for details.
	 */
	virtual void do_begin_session();
	virtual void do_end_session();
	virtual void do_flush();

	virtual void do_begin_transaction();
	virtual void do_commit_transaction();
	virtual void do_cancel_transaction();

	virtual om_docid do_add_document(const OmDocument & document);
	virtual void do_delete_document(om_docid did);
	virtual void do_replace_document(om_docid did,
					 const OmDocument & document);

	virtual void do_reopen();
	//@}

    public:

	/** Create and open a writable quartz database.
	 *
	 *  @exception OmOpeningError thrown if database can't be opened.
	 *
	 *  @param params Parameters supplied by the user to specify the
	 *                location of the database to open.
	 */
	QuartzWritableDatabase(const OmSettings & params);

	~QuartzWritableDatabase();

	/** Virtual methods of Database.
	 */
	//@{
	om_doccount  get_doccount() const;
	om_doclength get_avlength() const;
	om_doclength get_doclength(om_docid did) const;
	om_doccount get_termfreq(const om_termname & tname) const;
	om_termcount get_collection_freq(const om_termname & tname) const;
	bool term_exists(const om_termname & tname) const;

	LeafPostList * do_open_post_list(const om_termname & tname) const;
	LeafTermList * open_term_list(om_docid did) const;
	Document * open_document(om_docid did, bool lazy = false) const;
	AutoPtr<PositionList> open_position_list(om_docid did,
					  const om_termname & tname) const;
	TermList * open_allterms() const;
	//@}
};

#endif /* OM_HGUARD_QUARTZ_DATABASE_H */
